@@ -50,9 +50,10 @@
 
    In either case, the command itself, once string-ified, must be a
    JSON-formatted string with the aforementioned structure."
-  (:import [java.util.concurrent Semaphore])
+  (:import [java.util.concurrent Semaphore]
+           [org.joda.time DateTime])
   (:require [clojure.tools.logging :as log]
-            [puppetlabs.i18n.core :as i18n]
+            [puppetlabs.i18n.core :refer [trs]]
             [puppetlabs.puppetdb.scf.storage :as scf-storage]
             [puppetlabs.puppetdb.catalogs :as cat]
             [puppetlabs.puppetdb.reports :as report]
@@ -70,7 +71,7 @@
             [schema.core :as s]
             [puppetlabs.puppetdb.config :as conf]
             [puppetlabs.puppetdb.time :refer [to-timestamp]]
-            [clj-time.core :as time :refer [now]]
+            [clj-time.core :as time :refer [now interval in-millis]]
             [clj-time.format :as fmt-time]
             [clojure.set :as set]
             [clojure.core.async :as async]
@@ -135,9 +136,8 @@
 ;; * `:queue-time`: how long the message spent in the queue before processing
 ;; * `:retry-counts`: histogram containing the number of times
 ;;                    messages have been retried prior to suceeding
-;; * `:invalidated`: *CURRENTLY DISABLED* - see PDB-3108 commands marked as delete?,
-;;                   caused by a newer command was enqueued that will overwrite an
-;;                   existing one in the queue
+;; * `:invalidated`: commands marked as delete?, caused by a newer command
+;;                   was enqueued that will overwrite an existing one in the queue
 ;; * `:depth`: number of commands currently enqueued
 ;;
 
@@ -151,11 +151,7 @@
      :queue-time (histogram mq-metrics-registry (to-metric-name-fn :queue-time))
      :retry-counts (histogram mq-metrics-registry (to-metric-name-fn :retry-counts))
      :depth (counter mq-metrics-registry (to-metric-name-fn :depth))
-
-     ;; There's currently no way for a command to be invalidated due to a replication/HA issue that
-     ;; is to be addressed in PDB-3108.
-     ;; :invalidated (counter mq-metrics-registry (to-metric-name-fn :invalidated))
-
+     :invalidated (counter mq-metrics-registry (to-metric-name-fn :invalidated))
      :seen (meter mq-metrics-registry (to-metric-name-fn :seen))
      :size (histogram mq-metrics-registry (to-metric-name-fn :size))
      :processed (meter mq-metrics-registry (to-metric-name-fn :processed))
@@ -237,10 +233,12 @@
 
 (defmacro upon-error-throw-fatality
   [& body]
-  `(try+
+  `(try
     ~@body
-    (catch Throwable e#
-      (throw+ (fatality e#)))))
+    (catch Exception e1#
+      (throw+ (fatality e1#)))
+    (catch AssertionError e2#
+      (throw+ (fatality e2#)))))
 
 ;; ## Command submission
 
@@ -250,68 +248,73 @@
   [q
    command-chan
    ^Semaphore write-semaphore
-   command :- s/Str
-   version :- s/Int
-   certname :- s/Str
-   command-stream
-   command-callback]
+   {:keys [command certname command-stream compression] :as command-req} :- queue/command-req-schema]
   (try
     (.acquire write-semaphore)
     (time! (get @metrics :message-persistence-time)
-           (let [cmd (queue/store-command
-                       q command version certname command-stream command-callback)
+           (let [cmd (queue/store-command q command-req)
                  {:keys [id received]} cmd]
              (async/>!! command-chan cmd)
-             (log/debugf (i18n/trs "[%s-%d] '%s' command enqueued for %s")
-                         id (tcoerce/to-long received) command certname)))
+             (log/debug (trs "[{0}-{1}] ''{2}'' command enqueued for {3}"
+                             id
+                             (tcoerce/to-long received)
+                             command
+                             certname))))
     (finally
       (.release write-semaphore)
       (when command-stream
         (.close command-stream)))))
 
+(defn log-command-processed-messsage [id received-time start-time command-kw certname & [opts]]
+  ;; manually stringify these to avoid locale-specific formatting
+  (let [id (str id)
+        received-time (str (tcoerce/to-long received-time))
+        duration (str (in-millis (interval start-time (now))))
+        command-name (command-names command-kw)]
+    (if-let [{:keys [puppet-version]} opts]
+      (log/info (trs "[{0}-{1}] [{2} ms] ''{3}'' puppet v{4} command processed for {5}"
+                     id received-time duration command-name puppet-version certname))
+      (log/info (trs "[{0}-{1}] [{2} ms] ''{3}'' command processed for {4}"
+                     id received-time duration command-name certname)))))
+
 ;; Catalog replacement
 
 (defn replace-catalog*
-  [{:keys [certname version id received payload]} db]
+  [{:keys [certname version id received payload]} start-time db]
   (let [{producer-timestamp :producer_timestamp :as catalog} payload]
     (jdbc/with-transacted-connection' db :repeatable-read
       (scf-storage/maybe-activate-node! certname producer-timestamp)
       (scf-storage/replace-catalog! catalog received))
-    (log/infof (i18n/trs "[%s-%d] '%s' command processed for %s")
-               id (tcoerce/to-long received) (command-names :replace-catalog) certname)))
+    (log-command-processed-messsage id received start-time :replace-catalog certname)))
 
-(defn replace-catalog [{:keys [payload received version] :as command} db]
+(defn replace-catalog [{:keys [payload received version] :as command} start-time db]
   (let [validated-payload (upon-error-throw-fatality
                            (cat/parse-catalog payload version received))]
     (-> command
         (assoc :payload validated-payload)
-        (replace-catalog* db))))
+        (replace-catalog* start-time db))))
 
 ;; Fact replacement
 
 (defn replace-facts*
-  [{:keys [payload id received] :as command} db]
+  [{:keys [payload id received] :as command} start-time db facts-blacklist]
   (let [{:keys [certname values] :as fact-data} payload
-        producer-timestamp (:producer_timestamp fact-data)]
+        producer-timestamp (:producer_timestamp fact-data)
+        trimmed-facts (if (seq facts-blacklist)
+                        (update fact-data :values #(apply dissoc % facts-blacklist))
+                        fact-data)]
     (jdbc/with-transacted-connection' db :repeatable-read
       (scf-storage/maybe-activate-node! certname producer-timestamp)
-      (scf-storage/replace-facts! fact-data))
-    (log/infof (i18n/trs "[%s-%d] '%s' command processed for %s")
-               id (tcoerce/to-long received) (command-names :replace-facts) certname)))
+      (scf-storage/replace-facts! trimmed-facts))
+    (log-command-processed-messsage id received start-time :replace-facts certname)))
 
-(defn replace-facts [{:keys [payload version received] :as command} db]
-  (let [validated-payload (upon-error-throw-fatality
-                           (-> (case version
-                                 2 (fact/wire-v2->wire-v5 payload received)
-                                 3 (fact/wire-v3->wire-v5 payload)
-                                 4 (fact/wire-v4->wire-v5 payload)
-                                 payload)
-                               (update :values utils/stringify-keys)
-                               (update :producer_timestamp to-timestamp)
-                               (assoc :timestamp received)))]
-    (-> command
-        (assoc :payload validated-payload)
-        (replace-facts* db))))
+(defn replace-facts
+  [{:keys [payload version received] :as command} start-time db facts-blacklist]
+  (replace-facts* (upon-error-throw-fatality
+                    (assoc command :payload (fact/normalize-facts version received payload)))
+                  start-time
+                  db
+                  facts-blacklist))
 
 ;; Node deactivation
 
@@ -325,41 +328,36 @@
       deactivate-node-wire-v2->wire-3))
 
 (defn deactivate-node*
-  [{:keys [id received payload]} db]
+  [{:keys [id received payload]} start-time db]
   (let [certname (:certname payload)
         producer-timestamp (to-timestamp (:producer_timestamp payload (now)))]
     (jdbc/with-transacted-connection db
       (when-not (scf-storage/certname-exists? certname)
         (scf-storage/add-certname! certname))
       (scf-storage/deactivate-node! certname producer-timestamp))
-    (log/infof (i18n/trs "[%s-%d] '%s' command processed for %s")
-               id (tcoerce/to-long received) (command-names :deactivate-node) certname)))
+    (log-command-processed-messsage id received start-time :deactivate-node certname)))
 
-(defn deactivate-node [{:keys [payload version] :as command} db]
+(defn deactivate-node [{:keys [payload version] :as command} start-time db]
   (-> command
       (assoc :payload (case version
                         1 (deactivate-node-wire-v1->wire-3 payload)
                         2 (deactivate-node-wire-v2->wire-3 payload)
                         payload))
-      (deactivate-node* db)))
+      (deactivate-node* start-time db)))
 
 ;; Report submission
 
 (defn store-report*
-  [{:keys [payload id received]} db]
+  [{:keys [payload id received]} start-time db]
   (let [{:keys [certname puppet_version] :as report} payload
         producer-timestamp (to-timestamp (:producer_timestamp payload (now)))]
     (jdbc/with-transacted-connection db
       (scf-storage/maybe-activate-node! certname producer-timestamp)
       (scf-storage/add-report! report received))
-    (log/infof (i18n/trs "[%s-%d] '%s' puppet v%s command processed for %s")
-               id
-               (tcoerce/to-long received)
-               (command-names :store-report)
-               puppet_version
-               certname)))
+    (log-command-processed-messsage id received start-time :store-report certname
+                                    {:puppet-version puppet_version})))
 
-(defn store-report [{:keys [payload version received] :as command} db]
+(defn store-report [{:keys [payload version received] :as command} start-time db]
   (let [validated-payload (upon-error-throw-fatality
                            (s/validate report/report-wireformat-schema
                                        (case version
@@ -371,7 +369,7 @@
                                          payload)))]
     (-> command
         (assoc :payload validated-payload)
-        (store-report* db))))
+        (store-report* start-time db))))
 
 ;; ## Command processors
 
@@ -388,24 +386,25 @@
 
 (defn process-command!
   "Takes a command object and processes it to completion. Dispatch is
-  based on the command's name and version information"
-  [{command-name :command version :version delete? :delete? :as command} db]
+   based on the command's name and version information"
+  [{command-name :command version :version delete? :delete? :as command} db facts-blacklist]
   (when-not delete?
-    (condp supported-command-version? [command-name version]
-      "replace catalog" (replace-catalog command db)
-      "replace facts" (replace-facts command db)
-      "store report" (store-report command db)
-      "deactivate node" (deactivate-node command db))))
+    (let [start (now)]
+      (condp supported-command-version? [command-name version]
+        "replace catalog" (replace-catalog command start db)
+        "replace facts" (replace-facts command start db facts-blacklist)
+        "store report" (store-report command start db)
+        "deactivate node" (deactivate-node command start db)))))
 
 (defn warn-deprecated
   "Logs a deprecation warning message for the given `command` and `version`"
   [version command]
-  (log/warnf "command '%s' version %s is deprecated, use the latest version" command version))
+  (log/warn (trs "command ''{0}'' version {1} is deprecated, use the latest version" command version)))
 
 (defprotocol PuppetDBCommandDispatcher
   (enqueue-command
-    [this command version certname payload]
-    [this command version certname payload command-callback]
+    [this command version certname producer-ts payload compression]
+    [this command version certname producer-ts payload compression command-callback]
     "Submits the command for processing, and then returns its unique id.")
 
   (stats [this]
@@ -433,7 +432,9 @@
                   (catch Throwable e
                     (if (zero? n)
                       (throw e)
-                      (do (log/debug e (i18n/trs "Exception throw in L1 retry attempt {0}" (- (inc num-retries) n)))
+                      (do (log/debug e
+                                     (trs "Exception throw in L1 retry attempt {0}"
+                                          (- (inc num-retries) n)))
                           ::failure))))]
       (if (= result ::failure)
         (recur (dec n))
@@ -441,11 +442,12 @@
 
 (def quick-retry-count 4)
 
-(defn process-command-and-respond! [{:keys [callback] :as cmd} db response-pub-chan stats-atom]
+(defn process-command-and-respond!
+  [{:keys [callback] :as cmd} db response-pub-chan stats-atom facts-blacklist]
   (try
     (let [result (call-with-quick-retry quick-retry-count
                   (fn []
-                    (process-command! cmd db)))]
+                    (process-command! cmd db facts-blacklist)))]
       (swap! stats-atom update :executed-commands inc)
       (callback {:command cmd :result result})
       (async/>!! response-pub-chan
@@ -474,11 +476,7 @@
   (mark-both-metrics! (:command cmdref) (:version cmdref) :discarded))
 
 (defn process-delete-cmdref
-  "Note that currently this function is unreachable, because commands
-  cannot currently be invalidated due to a replication/HA issue that
-  is to be addressed in PDB-3108.
-
-  Processes a command ref marked for deletion. This is similar to
+  "Processes a command ref marked for deletion. This is similar to
   processing a non-delete cmdref except different metrics need to be
   updated to indicate the difference in command"
   [{:keys [command version] :as cmdref} q scf-write-db response-chan stats]
@@ -491,25 +489,32 @@
   "Parses, processes and acknowledges a successful command ref and
   updates the relevant metrics. Any exceptions that arise are
   unhandled and expected to be caught by the caller."
-  [cmdref q scf-write-db response-chan stats]
+  [cmdref q scf-write-db response-chan stats facts-blacklist]
   (let [{:keys [command version] :as cmd} (queue/cmdref->cmd q cmdref)
         retries (count (:attempts cmdref))]
+    (if-not cmd
+      ;; Queue file is missing
+      (let [{:keys [command version]} cmdref]
+        (create-metrics-for-command! command version)
+        (mark-both-metrics! command version :seen)
+        (update-counter! :depth command version dec!)
+        (mark! (global-metric :fatal)))
+      (do
+        (create-metrics-for-command! command version)
 
-    (create-metrics-for-command! command version)
+        (mark-both-metrics! command version :seen)
+        (update! (global-metric :retry-counts) retries)
+        (update! (cmd-metric command version :retry-counts) retries)
 
-    (mark-both-metrics! command version :seen)
-    (update! (global-metric :retry-counts) retries)
-    (update! (cmd-metric command version :retry-counts) retries)
+        (mutils/multitime!
+         [(global-metric :processing-time)
+          (cmd-metric command version :processing-time)]
 
-    (mutils/multitime!
-     [(global-metric :processing-time)
-      (cmd-metric command version :processing-time)]
+         (process-command-and-respond! cmd scf-write-db response-chan stats facts-blacklist)
+         (mark-both-metrics! command version :processed))
 
-     (process-command-and-respond! cmd scf-write-db response-chan stats)
-     (mark-both-metrics! command version :processed))
-
-    (queue/ack-command q cmd)
-    (update-counter! :depth command version dec!)))
+        (queue/ack-command q cmd)
+        (update-counter! :depth command version dec!)))))
 
 (def command-delay-ms (* 1000 60 60))
 
@@ -535,7 +540,7 @@
   "Processes the message via (process-message msg), retrying messages
   that fail via (delay-message msg), and discarding messages that have
   fatal errors or have exceeded their maximum allowed attempts."
-  [q command-chan dlo delay-pool scf-write-db response-chan stats]
+  [q command-chan dlo delay-pool scf-write-db response-chan stats facts-blacklist]
   (fn [{:keys [certname command version received delete? id] :as cmdref}]
     (try+
 
@@ -551,34 +556,34 @@
        (process-delete-cmdref cmdref q scf-write-db response-chan stats)
        (let [retries (count (:attempts cmdref))]
          (try+
-          (process-cmdref cmdref q scf-write-db response-chan stats)
+          (process-cmdref cmdref q scf-write-db response-chan stats facts-blacklist)
           (catch fatal? obj
             (mark! (global-metric :fatal))
             (let [ex (:cause obj)]
               (log/error
                (:wrapper &throw-context)
-               (i18n/trs "[{0}] [{1}] Fatal error on attempt {2} for {3}" id command retries certname))
+               (trs "[{0}] [{1}] Fatal error on attempt {2} for {3}" id command retries certname))
               (discard-message cmdref ex q dlo)))
           (catch Exception _
             (let [ex (:throwable &throw-context)]
               (mark-both-metrics! command version :retried)
               (if (< retries maximum-allowable-retries)
                 (do
-                  (log/errorf
+                  (log/error
                    ex
-                   (i18n/trs "[{0}] [{1}] Retrying after attempt {2} for {3}, due to: {4}"
-                             id command retries certname ex))
+                   (trs "[{0}] [{1}] Retrying after attempt {2} for {3}, due to: {4}"
+                        id command retries certname ex))
                   (send-delayed-message cmdref ex command-chan delay-pool))
                 (do
                   (log/error
                    ex
-                   (i18n/trs "[{0}] [{1}] Exceeded max attempts ({2}) for {3}" id command retries certname))
+                   (trs "[{0}] [{1}] Exceeded max attempts ({2}) for {3}" id command retries certname))
                   (discard-message cmdref ex q dlo))))))))
 
      (catch [:kind ::queue/parse-error] _
        (mark! (global-metric :fatal))
        (log/error (:wrapper &throw-context)
-                  (i18n/trs "Fatal error parsing command: {0}" (:id cmdref)))
+                  (trs "Fatal error parsing command: {0}" (:id cmdref)))
        (discard-message cmdref (:throwable &throw-context) q dlo)))))
 
 (defn create-command-handler-threadpool
@@ -624,7 +629,8 @@
                                       delay-pool
                                       scf-write-db
                                       response-chan
-                                      (:stats context))]
+                                      (:stats context)
+                                      (get-in (get-config) [:database :facts-blacklist]))]
 
       (doto (Thread. (fn []
                        (try+ (gtp/dochan command-threadpool handle-cmd command-chan)
@@ -659,17 +665,17 @@
   (stats [this]
     @(:stats (service-context this)))
 
-  (enqueue-command [this command version certname command-stream]
-                   (enqueue-command this command version certname command-stream identity))
+  (enqueue-command [this command version certname producer-ts command-stream compression]
+                   (enqueue-command this command version certname producer-ts command-stream compression identity))
 
-  (enqueue-command [this command version certname command-stream command-callback]
+  (enqueue-command [this command version certname producer-ts command-stream compression command-callback]
     (let [config (get-config)
           q (:q (shared-globals))
           command-chan (:command-chan (shared-globals))
           write-semaphore (:write-semaphore (service-context this))
           command (if (string? command) command (command-names command))
-          result (do-enqueue-command q command-chan write-semaphore command
-                                     version certname command-stream command-callback)]
+          command-req (queue/create-command-req command version certname producer-ts compression command-callback command-stream)
+          result (do-enqueue-command q command-chan write-semaphore command-req)]
       ;; Obviously assumes that if do-* doesn't throw, msg is in
       (inc-cmd-depth command version)
       (swap! (:stats (service-context this)) update :received-commands inc)
